package cn.dankal.share.cache.factory;

import cn.hutool.core.util.StrUtil;

import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * ThreadExecutorFactory
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2020-12-23
 */
public class ThreadExecutorFactory {


    /**
     * 获取  ThreadPoolExecutor
     *
     * @param corePoolSize       核心线程数
     * @param maximumPoolSize    最大线程数
     * @param keepAliveTime      超过核心线程数的线程存活时间
     * @param timeUnit           超过核心线程数的线程存活时间单位
     * @param threadNamePrefix   创建线程的名称前缀
     * @param blockQueueCapacity 线程池阻塞队列的容量
     * @return ThreadPoolExecutor
     * @see ThreadPoolExecutor
     */
    public static ThreadPoolExecutor getThreadPoolExecutor(int corePoolSize,
                                                           int maximumPoolSize,
                                                           long keepAliveTime,
                                                           TimeUnit timeUnit,
                                                           String threadNamePrefix,
                                                           int blockQueueCapacity) {
        // 参数校验
        if (corePoolSize <= 0) {
            throw new IllegalArgumentException("corePoolSize 必须大于0");
        }
        if (maximumPoolSize <= 0) {
            throw new IllegalArgumentException("maximumPoolSize 必须大于0");
        }
        if (keepAliveTime < 0) {
            throw new IllegalArgumentException("keepAliveTime 必须大于0");
        }
        if (null == timeUnit) {
            throw new IllegalArgumentException("timeUnit 不能为空");
        }
        if (corePoolSize > maximumPoolSize) {
            throw new IllegalArgumentException("corePoolSize 必须小于等于 maximumPoolSize");
        }
        if (StrUtil.isBlank(threadNamePrefix)) {
            throw new IllegalArgumentException("threadNamePrefix 不能为空");
        }
        if (blockQueueCapacity <= 0) {
            throw new IllegalArgumentException("blockQueueCapacity 必须大于0");
        }

        return new ThreadPoolExecutor(
                corePoolSize,
                maximumPoolSize,
                keepAliveTime,
                timeUnit,
                new ArrayBlockingQueue<>(blockQueueCapacity),
                threadFactory(threadNamePrefix)
        );
    }


    /**
     * 初始化 ThreadFactory 并为 创建的 线程 命名
     *
     * @param namePrefix 线程名称前缀
     * @return ThreadFactory
     */
    private static ThreadFactory threadFactory(String namePrefix) {
        return (runnable -> {
            Thread thread = new Thread(runnable);
            long id = thread.getId();
            thread.setName(namePrefix + "-" + id);
            return thread;
        });
    }

    /**
     * 获取 一个  指定 {@code maximumPoolSize} 的  {@link ThreadPoolExecutor}
     * <pre>
     *    corePoolSize: 1
     *    keepAliveTime: 300 ms
     *    blockQueueCapacity: 100
     *  </pre>
     *
     * @param namePrefix      namePrefix
     * @param maximumPoolSize maximumPoolSize
     * @return {@link ThreadPoolExecutor}
     */
    public static ThreadPoolExecutor getThreadPoolExecutor(String namePrefix, Integer maximumPoolSize) {
        return getThreadPoolExecutor(
                1,
                maximumPoolSize,
                300,
                TimeUnit.MILLISECONDS,
                namePrefix,
                100
        );
    }
}
