package cn.dankal.share.cache.reference.config;

import cn.hutool.core.util.StrUtil;
import org.redisson.Redisson;
import org.redisson.api.RedissonClient;
import org.redisson.config.Config;
import org.springframework.boot.autoconfigure.data.redis.RedisProperties;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.util.CollectionUtils;

import java.util.List;

/**
 * RedissonBaseConfig
 * <br/>
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021/02/26
 * @copyright 蛋壳创意科技 - www.dankal.cn
 */
@Configuration
public class RedissonBaseConfig {

    /**
     * RedisProperties
     */
    private final RedisProperties redisProperties;

    /**
     * redisson  redis host  prefix
     */
    private final static String ADDRESS_PREFIX = "redis://";

    public RedissonBaseConfig(RedisProperties redisProperties) {
        this.redisProperties = redisProperties;
    }

    @Bean(destroyMethod ="shutdown")
    public RedissonClient redissonClient() {
        Config config = new Config();
        // 验证是否配置了 redis  host/ redis nodes
        String redisHost = redisProperties.getHost();
        RedisProperties.Cluster propertiesCluster = redisProperties.getCluster();
        boolean checkRedisHost = StrUtil.isBlank(redisHost) && (propertiesCluster == null || CollectionUtils.isEmpty(propertiesCluster.getNodes()));
        if (checkRedisHost) {
            throw new IllegalArgumentException("需要配置 redis.host 或  redis.cluster.nodes");
        }
        if (null != propertiesCluster && !CollectionUtils.isEmpty(propertiesCluster.getNodes())) {
            // redis  cluster
            config.useClusterServers()
                    .addNodeAddress(this.transferRedissonRedisNode(propertiesCluster.getNodes()))
                    .setPassword(redisProperties.getPassword())
                    .setTimeout((int) redisProperties.getTimeout().toMillis());
        } else {
            // redis singleton
            config.useSingleServer()
                    .setAddress(ADDRESS_PREFIX + redisHost + ":" + redisProperties.getPort())
                    .setTimeout((int) redisProperties.getTimeout().toMillis())
                    .setPassword(redisProperties.getPassword());
        }
        return Redisson.create(config);
    }

    /**
     * 将 node list 转化为  redisson redis cluster config
     * <p>
     * 127.0.0.0:6379  --> redis://127.0.0.1:6379
     * </p>
     *
     * @param redisNodeList redisNodeList
     * @return string []
     */
    private String[] transferRedissonRedisNode(List<String> redisNodeList) {
        String[] resultArray = new String[redisNodeList.size()];
        for (int i = 0; i < resultArray.length; i++) {
            resultArray[i] = ADDRESS_PREFIX + redisNodeList.get(i);
        }
        return resultArray;
    }
}
