package cn.dankal.share.cache.service.impl;

import cn.dankal.share.cache.common.Const;
import cn.dankal.share.cache.entity.Provinces;
import cn.dankal.share.cache.mapper.provinces.ProvincesMapper;
import cn.dankal.share.cache.service.IProvincesService;
import cn.hutool.bloomfilter.BitMapBloomFilter;
import cn.hutool.core.bean.BeanUtil;
import cn.hutool.core.util.RandomUtil;
import cn.hutool.core.util.StrUtil;
import com.baomidou.mybatisplus.core.conditions.Wrapper;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.mapper.BaseMapper;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.google.gson.Gson;
import com.google.gson.reflect.TypeToken;
import lombok.extern.slf4j.Slf4j;
import org.springframework.cache.annotation.CacheEvict;
import org.springframework.cache.annotation.Cacheable;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.stereotype.Service;

import javax.annotation.PostConstruct;
import java.util.Objects;
import java.util.Optional;
import java.util.concurrent.TimeUnit;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-03-26
 */
@Slf4j
@Service
public class IProvincesServiceImpl extends ServiceImpl<ProvincesMapper, Provinces> implements IProvincesService {


    /**
     * gson
     */
    private final Gson gson;

    /**
     * stringRedisTemplate
     */
    private final StringRedisTemplate stringRedisTemplate;

    private BitMapBloomFilter bitMapBloomFilter;

    public IProvincesServiceImpl(Gson gson, StringRedisTemplate stringRedisTemplate) {
        this.gson = gson;
        this.stringRedisTemplate = stringRedisTemplate;
    }

    @PostConstruct
    public void postConstruct() {
        bitMapBloomFilter = new BitMapBloomFilter(10);
        LambdaQueryWrapper<Provinces> provincesLambdaQueryWrapper = new LambdaQueryWrapper<Provinces>()
                .select(Provinces::getProvinceId);
        baseMapper.selectList(provincesLambdaQueryWrapper)
                .stream()
                .filter(Objects::nonNull)
                .map(Provinces::getProvinceId)
                .forEach(bitMapBloomFilter::add);
    }

    @Override
    public Optional<Provinces> detailNoCache(String provincesId) {
        if (StrUtil.isBlank(provincesId)) {
            return Optional.empty();
        }
        Provinces provinces = getProvincesByProvinces(provincesId);
        if (log.isDebugEnabled()) {
            log.debug("detailNoCache ==> {} get Provinces by {} from database.", Thread.currentThread().getName(), provincesId);
        }
        return Optional.ofNullable(provinces);
    }

    /**
     * 通过 provincesId 从数据库中查询 {@link Provinces}
     *
     * @param provincesId provincesId
     * @return Provinces
     * @see {@link BaseMapper#selectOne(Wrapper)}
     */
    private Provinces getProvincesByProvinces(String provincesId) {
        if (log.isDebugEnabled()) {
            log.debug("getProvincesByProvinces ==> get {} from database.", provincesId);
        }
        LambdaQueryWrapper<Provinces> provincesLambdaQueryWrapper = new LambdaQueryWrapper<Provinces>()
                .eq(Provinces::getProvinceId, provincesId);
        return baseMapper.selectOne(provincesLambdaQueryWrapper);
    }

    @Override
    public Optional<Provinces> detailWithRedisCache(String provincesId) {
        if (StrUtil.isBlank(provincesId)) {
            return Optional.empty();
        }
        String key = StrUtil.format(Const.RedisKey.KEY_PATTERN_PROVINCES_DETAIL, provincesId);
        String cache = stringRedisTemplate.opsForValue().get(key);
        if (StrUtil.isNotBlank(cache)) {
            if (log.isDebugEnabled()) {
                log.debug("detailWithRedisCache  ==> {} success hit cache {}", Thread.currentThread().getName(), cache);
            }
            // 缓存命中
            return gson.fromJson(cache, new TypeToken<Optional<Provinces>>() {
            }.getType());
        } else {
            // 数据没有缓存
            // TODO 缓存击穿 分布式锁  etc. Redisson 可以参阅 cn.dankal.share.cache.reference
            Provinces provinces = this.getProvincesByProvinces(provincesId);
            Optional<Provinces> provincesOptional;
            if (BeanUtil.isNotEmpty(provinces)) {
                if (log.isDebugEnabled()) {
                    log.debug("detailWithRedisCache  ==> {} success get provinces by {} from  database.", Thread.currentThread().getName(), provincesId);
                }
                provincesOptional = Optional.of(provinces);
                final long timeoutInMinutes = Const.CacheTimeout.TIMEOUT_PROVINCES_DETAIL_MINUTES;
                stringRedisTemplate.opsForValue().set(key, gson.toJson(provincesOptional), timeoutInMinutes, TimeUnit.MINUTES);
            } else {
                // 防止缓存穿透
                if (log.isDebugEnabled()) {
                    log.debug("detailWithRedisCache  ==> {} preven cache break down .", Thread.currentThread().getName());
                }
                final long timeoutInMinutes = Const.CacheTimeout.TIMEOUT_PROVINCES_DETAIL_MINUTES;
                provincesOptional = Optional.empty();
                stringRedisTemplate.opsForValue()
                        .set(key, gson.toJson(provincesOptional), timeoutInMinutes, TimeUnit.MINUTES);
            }
            // 防止缓存雪崩
            // 每一个缓存 30秒 以内的随机差距
            Long expire = stringRedisTemplate.getExpire(key, TimeUnit.SECONDS);
            expire += RandomUtil.randomLong(30);
            stringRedisTemplate.expire(key, expire, TimeUnit.SECONDS);
            return provincesOptional;
        }
    }


    @CacheEvict(cacheNames = "provinces_detail_cache", key = "#provinces.provinceId")

    @Override
    public Optional<Integer> updateProvinces(Provinces provinces) {
        LambdaQueryWrapper<Provinces> provincesLambdaQueryWrapper = new LambdaQueryWrapper<Provinces>()
                .eq(Provinces::getProvinceId, provinces.getProvinceId());
        int update = baseMapper.update(provinces, provincesLambdaQueryWrapper);
        return Optional.of(update);
    }

    @Cacheable(cacheNames = "provinces_detail_cache", keyGenerator = "customerKeyGenerator")
    @Override
    public Optional<Provinces> detailWithSpringCache(String provincesId) {
        Provinces provinces = this.getProvincesByProvinces(provincesId);
        return Optional.ofNullable(provinces);
    }

    @Override
    public Optional<Provinces> detailsWithBoolmFilter(String provinceId) {
        if (StrUtil.isBlank(provinceId)) {
            return Optional.empty();
        }
        if (!bitMapBloomFilter.contains(provinceId)) {
            // provinceId 不存在
            return Optional.empty();
        }
        return Optional.ofNullable(this.getProvincesByProvinces(provinceId));
    }
}
